open Block;;
open Parser;;
open Printf;;

(* attributes stuff *)

type emphasis = On | Off | Strong;;
type color = Color of char | NoColor
type size = Size of int | NoSize
type attributes = { b: bool; i: bool; u: int; em: emphasis; color: color;
tt: bool; size: size; }

let initial_attributes = { b = false; i = false; u = 0; em = Off; color = NoColor;
tt = false; size = NoSize; }

let string_of_em em = match em with | On -> "E" | Off -> "e" | Strong -> "S"
let string_of_size s = match s with   Size n -> string_of_int n | NoSize -> "-"
let string_of_color c = match c with Color c -> String.make 1 c | NoColor -> "-"

let boolchar c b =
  if b then String.get (String.uppercase (String.make 1 c)) 0 else c

let string_of_attributes a =
  sprintf "%c%c%c%s%d%s%s" (boolchar 'b' a.b) (boolchar 'i' a.i)
    (boolchar 't' a.tt) (string_of_em a.em) a.u (string_of_color a.color)
    (string_of_size a.size)

let spaceattributes_for_attributes c =
  { b = false; i = false; u = c.u; em = Off;
    color = (if c.u == 0 then Color 'w' else c.color);
    tt = c.tt; size = c.size }

let tagB c = { b = true; i = c.i; u = c.u; em = c.em; color = c.color; tt = c.tt;
               size = c.size; }
let tagS c = { b = c.b; i = c.i; u = c.u; em = Strong; color = c.color; tt = c.tt;
               size = c.size; }
let tagEM c =
  let v =
    if c.em == On then Off
    else if c.em == Off then On
    else Strong
  in { b = c.b; i = c.i; u = c.u; em = v;
       color = c.color; tt = c.tt; size = c.size; }

let tagI c = { b = c.b; i = true; u = c.u; em = c.em; color = c.color; tt = c.tt;
               size = c.size; }
let tagTT c = { b = c.b; i = c.i; u = c.u; em = c.em; color = c.color; tt = true;
               size = c.size; }
let tagU c = 
  let v =
    if c.u < 3 then c.u + 1
    else c.u
  in { b = c.b; i = c.i; u = v; em = c.em;
       color = c.color; tt = c.tt; size = c.size; }

let tagPL c =
  { b = false; i = false; u = 0; em = Off; color = c.color; tt = false;
    size = c.size; }

let tagNumber n c =
  { b = c.b; i = c.i; u = c.u; em = c.em; color = c.color;
    tt = c.tt; size = Size n; }

let tagColor col c =
  { b = c.b; i = c.i; u = c.u; em = c.em;
    color = Color col; tt = c.tt; size = c.size; }

let tag_table = Hashtbl.create 53
  let _ =
    List.iter (fun (tag, f) -> Hashtbl.add tag_table tag f)
              [ 'B', tagB; 'S', tagS; 'E', tagEM; 'I', tagI;
		'T', tagTT; 'P', tagPL; 'U', tagU;
		'0', tagNumber 0; '1', tagNumber 1;
		'2', tagNumber 2; '3', tagNumber 3;
		'4', tagNumber 4; '5', tagNumber 5;
		'6', tagNumber 6; '7', tagNumber 7;
		'8', tagNumber 8; '9', tagNumber 9;
		'r', tagColor 'r'; 'g', tagColor 'g';
		'b', tagColor 'b'; 'c', tagColor 'c';
		'm', tagColor 'm'; 'y', tagColor 'y';
		'k', tagColor 'k'; 'w', tagColor 'w';]

let applytagtoattrib tag attrib =
  let f = Hashtbl.find tag_table tag in f attrib

let b_escaped s =
  let e = (String.escaped s) in
  for i = 0 to (String.length e)-1 do
    if String.get e i == ' ' then String.set e i '.'
  done; e

(* Create decorated output *)

let is_whitespace s =
  String.length s == 0 || let c = String.get s 0 in match c with
  | '\n' -> true  | '\t' -> true  | '\r' -> true  | ' ' -> true
  | _ -> false
let collapse_if_not_tt s attribs =
  if attribs.tt || not (is_whitespace s) then s
  else "\n"

let rec generate_decorated_output attribs list = match list with
| [] -> []
| Text t :: tl -> (attribs, t) :: (generate_decorated_output attribs tl)
| Space s :: tl ->
    let s_attribs = spaceattributes_for_attributes attribs in
    let cs = collapse_if_not_tt s attribs in
    let new_node = (s_attribs, cs)
    and rest = (generate_decorated_output attribs tl) in
    if List.length rest > 0 then
      let (n_attribs, n_text) = List.hd rest in
      if is_whitespace n_text && n_attribs = s_attribs && (not s_attribs.tt) then
	rest
      else
	new_node :: rest
    else
      new_node :: rest
| Block (tag,bl) :: tl ->
    (* BRAIN DAMAGE -- joining lists can crunch spaces together *)
    (generate_decorated_output (applytagtoattrib tag attribs) bl) @
    (generate_decorated_output attribs tl)

(* this one doesn't remove adjacent whitespace pairs *)
let rec generate_decorated_output_raw attribs list = match list with
| [] -> []
| Text t :: tl -> (attribs, t) :: (generate_decorated_output_raw attribs tl)
| Space s :: tl ->
 (spaceattributes_for_attributes attribs, collapse_if_not_tt s attribs) ::
    (generate_decorated_output_raw attribs tl)
| Block (tag,bl) :: tl ->
    (generate_decorated_output_raw (applytagtoattrib tag attribs) bl) @
    (generate_decorated_output_raw attribs tl)

let rec print_decorated_output list = match list with
| [] -> ()
| (attribs, t) :: tl -> printf "%s : %s\n" (string_of_attributes attribs) (b_escaped t);
    print_decorated_output tl

let po r = print_decorated_output (generate_decorated_output initial_attributes r);;
let por r = print_decorated_output (generate_decorated_output_raw initial_attributes r);;


(* returns (l1 leftover, b1 leftover, l2 leftover, boolean) *)
let rec dco_prefix_match (l1, b1, att1) (l2, b2, att2) =
(* if both l1 & l2 are not empty, bite one off *)
(* if one is empty, assume it is l2 (swap otherwise) *)
(* build up l2; consume a piece of b2 *)
(* only return when (1) mismatch (2) b2 is empty *)
  match l2 with
  | hd2 :: tl2 ->
      (match l1 with
      |	hd1 :: tl1 -> 
	  if hd1 = hd2 then
	    dco_prefix_match (tl1, b1, att1) (tl2, b2, att2)
	  else
	    (l1, b1, l2, b2, false)
      |	_ ->
	  let (l2, b2, l1, b1, ok) =
	    dco_prefix_match (l2, b2, att2) (l1, b1, att1) in
(* Finish consuming b1 *)
	  (l1, b1, l2 @ generate_decorated_output_raw att2 b2, [], ok))
  | [] -> (* build up some l2 *)
      match b2 with
      | [] -> (l1, b1, l2, b2, true) (* no more b2 to consume! *)
      | Text t2 :: tl2 -> 
	  dco_prefix_match (l1, b1, att1) ([att2, t2], tl2, att2)
      | Space s2 :: tl2 ->
	  let sp_att2 = spaceattributes_for_attributes att2
	  and s2 = collapse_if_not_tt s2 att2 in
	  dco_prefix_match (l1, b1, att1) ([sp_att2, s2], tl2, att2)
      | Block (t2, bl2) :: tl2 ->
	  let n_attrib = (applytagtoattrib t2 att2) in
	  let n_l1, b1, n_l2, b2, ok =
	    dco_prefix_match (l1, b1, att1) (l2, bl2, n_attrib) in
	  if ok then dco_prefix_match (n_l1, b1, att1) (n_l2, tl2, att2)
	  else (n_l1, b1, n_l2, b2, ok)

let decorated_output_confirmed_match (b1, att1) (b2, att2) =
  let l1, b1, l2, b2, ok = dco_prefix_match ([], b1, att1) ([], b2, att2) in 
  ok && (List.length l1 = 0) && (List.length l2 = 0)
;;
(*# trace dco_prefix_match;;
# trace decorated_output_confirmed_match;;*)

(* optimize out useless tags *)
(* This algorithm goes through, and whenever it finds a tag, it sees
   "what would happen" if that tag were to be removed. If the output
   is still the same, the tag is discarded. *)

let rec remove_useless_blocks attribs list = match list with
| [] -> []
| Block (tag,bl) :: tl ->
    (* if we dispense of this tag, does the subblock look the same? *)
    let n_attribs = (applytagtoattrib tag attribs) in
    if decorated_output_confirmed_match (bl, attribs) (bl, n_attribs) then
      (remove_useless_blocks attribs bl) @ (remove_useless_blocks attribs tl)
    else
      Block (tag, remove_useless_blocks n_attribs bl) :: (remove_useless_blocks attribs tl)
| hd :: tl -> hd :: remove_useless_blocks attribs tl

let remove_useless = remove_useless_blocks initial_attributes

(* collapse whitespace *)
(* This algorithm goes through and replaces all multi-character whitespace *)
(* blocks with single whitespace blocks (unless TT is in effect) *)

let space_replace s = String.contains s '\n'
let rec collapse_whitespace_blocks attribs list = match list with
| [] -> [] 
| Block (tag,bl) :: tl -> 
    let n_attribs = (applytagtoattrib tag attribs) in  
    Block (tag, collapse_whitespace_blocks n_attribs bl) ::
    (collapse_whitespace_blocks attribs tl)
| Space s :: tl ->
    if attribs.tt then Space s :: (collapse_whitespace_blocks attribs tl)
    else Space (if String.contains s '\n' then "\n" else " ") ::
      (collapse_whitespace_blocks attribs tl)
| Text t :: tl -> Text t :: (collapse_whitespace_blocks attribs tl)

let collapse_whitespace = collapse_whitespace_blocks initial_attributes

(* merge whitespace *)
(* This algorithm goes through and merges adjacent similar whitespace blocks *)
(* returns (LIST, last_space_attribs) *)

type last_space_attribs = Attribs of attributes | NoAttribs;;

let rec merge_whitespace_blocks attribs list ls_attribs = match list with
| [] -> ([], ls_attribs)
| Block (tag,bl) :: tl -> 
    let n_attribs = (applytagtoattrib tag attribs) in
    let (bl, ls_attribs) = merge_whitespace_blocks n_attribs bl ls_attribs in
    let (tl, ls_attribs) = merge_whitespace_blocks attribs tl ls_attribs in
    (Block (tag, bl) :: tl, ls_attribs)
| Space s :: tl ->
    let sa_attribs = (spaceattributes_for_attributes attribs) in
    if attribs.tt || (not ((Attribs sa_attribs) = ls_attribs)) then
      let ls_attribs = (if attribs.tt then NoAttribs else Attribs sa_attribs) in
      let (tl, ls_attribs) = merge_whitespace_blocks attribs tl ls_attribs in
      (Space s :: tl, ls_attribs)
    else
      merge_whitespace_blocks attribs tl ls_attribs
| Text t :: tl ->
      let (tl, ls_attribs) = merge_whitespace_blocks attribs tl NoAttribs in
      (Text t :: tl, ls_attribs)

let merge_whitespace list =
  fst(merge_whitespace_blocks initial_attributes list NoAttribs)

(* remove <EM><EM>...</EM></EM> pattern *)

let rec remove_em_pairs list = match list with
| [] -> []
| Block ('E',[ Block ('E', bl)]) :: tl ->
    (* we have the pattern! *)
    (remove_em_pairs bl) @ (remove_em_pairs tl)
| Block (t, bl) :: tl -> Block (t, remove_em_pairs bl) :: remove_em_pairs tl
| hd :: tl -> hd :: remove_em_pairs tl

(* sort blocks: we want <col><siz><B><I><U><TT><S><EM> *)
(* If we see <tag2><tag1>...</tag1></tag2> replace with
   <tag1><tag2>...</tag2></tag1> *)

let is_color_tag t = match t with
|'r'|'g'|'b'|'c'|'m'|'y'|'k'|'w' -> true | _ -> false

let is_size_tag t = match t with
|'0'|'1'|'2'|'3'|'4'|'5'|'6'|'7'|'8'|'9' -> true | _ -> false

let do_tags_commute t1 t2 =
  let att = initial_attributes in
  (applytagtoattrib t1 (applytagtoattrib t2 att)) =
  (applytagtoattrib t2 (applytagtoattrib t1 att));;

let tag_priority t = match t with
|'r'|'g'|'b'|'c'|'m'|'y'|'k'|'w' -> 0
|'0'|'1'|'2'|'3'|'4'|'5'|'6'|'7'|'8'|'9' -> 1
| 'P' -> 2
| 'B' -> 3
| 'I' -> 4
| 'U' -> 5
| 'E' -> 6
| 'S' -> 7
| 'T' -> 8
| _ -> 9

let should_swap t1 t2 =
  do_tags_commute t1 t2 && tag_priority t1 > tag_priority t2

let rec sort_tags list = match list with
| [] -> []
| Block (tag1,[ Block (tag2, bl)]) :: tl ->
    if should_swap tag1 tag2 then
      sort_tags (Block (tag2, sort_tags [ Block (tag1, bl)]) :: sort_tags tl)
    else
      Block (tag1, sort_tags [ Block (tag2, bl)]) :: sort_tags tl
| Block (t, bl) :: tl -> Block (t, sort_tags bl) :: sort_tags tl
| hd :: tl -> hd :: sort_tags tl

(* bring spaces inward (or outward) *)
(* try <tag1> <tag2>foo</tag2></tag1> -> <tag1><tag2> foo</tag2></tag1> *)
(* or  <tag1> <tag2>foo</tag2></tag1> ->  <tag1><tag2>foo</tag2></tag1>  -- TURNED OFF TO PREVENT LOOPING *)
(* Only do it if the space style will be the same *)
(* This should allow more tag commutation *)

let rec sort_spaces_attr attribs list = match list with
| [] -> []
| Block (tag1,[ Space s; Block (tag2, bl)]) :: tl ->
    (* can we commute the space inward or outward? *)
    let att1 = attribs in
    let att2 = (applytagtoattrib tag1 att1) in
    let att3 = (applytagtoattrib tag2 att2) in
    let s_att1 = spaceattributes_for_attributes att1
    and s_att2 = spaceattributes_for_attributes att2
    and s_att3 = spaceattributes_for_attributes att3
    and tl = sort_spaces_attr att1 tl in
    if s_att2 = s_att3 then
      Block (tag1, sort_spaces_attr att2 [ Block (tag2, Space s :: bl)]) :: tl
(*    else if s_att2 = s_att1 then
      Space s :: sort_spaces_attr att1 (Block (tag1, [ Block (tag2, bl)]) :: tl)
*)
    else
      Block (tag1,[ Space s; Block (tag2, sort_spaces_attr att3 bl)]) :: tl
| Block (t, bl) :: tl ->
    let n_attribs = (applytagtoattrib t attribs) in
    Block (t, sort_spaces_attr n_attribs bl) :: sort_spaces_attr attribs tl
| hd :: tl -> hd :: sort_spaces_attr attribs tl
let sort_spaces = sort_spaces_attr initial_attributes

(* Change <tag1>foo</tag1><tag2>bar</tag2> to *)
(* <tag2><tag1>foo</tag1>bar</tag2> *)
(* when tags commute and they are equivalent *)

let rec embed_tags_attr attribs list = match list with
| [] -> []
| Block (tag1, bl1) :: Block(tag2, bl2) :: tl ->
    (* can we embed? *)
    let original = [Block (tag1, bl1); Block(tag2, bl2)]
    and trial = [Block (tag2, Block(tag1, bl1) :: bl2)] in
    let ok =
      do_tags_commute tag1 tag2 &&
      decorated_output_confirmed_match (original, attribs) (trial, attribs) in
    if ok then
      embed_tags_attr attribs trial @ tl
    else
      Block (tag1, embed_tags_attr (applytagtoattrib tag1 attribs) bl1) ::
      embed_tags_attr attribs (Block(tag2, bl2) :: tl)
| Block (t, bl) :: tl ->
    let n_attribs = (applytagtoattrib t attribs) in
    Block (t, embed_tags_attr n_attribs bl) :: embed_tags_attr attribs tl
| hd :: tl -> hd :: embed_tags_attr attribs tl
let embed_tags = embed_tags_attr initial_attributes


(* Change <tag1>foo</tag1><tag2>bar</tag2> to *)
(* <tag1>foo <tag2>bar</tag2></tag1> *)
(* when tags commute and they are equivalent *)

let rec embed_tags_end_attr attribs list = match list with
| [] -> []
| Block (tag1, bl1) :: Block(tag2, bl2) :: tl ->
    (* can we embed? *)
    let original = [Block (tag1, bl1); Block(tag2, bl2)]
    and trial = [Block (tag1, bl1 @ [Block(tag2, bl2)])] in
    let ok =
      do_tags_commute tag1 tag2 &&
      decorated_output_confirmed_match (original, attribs) (trial, attribs) in
    if ok then
      embed_tags_end_attr attribs trial @ tl
    else
      Block (tag1, embed_tags_end_attr (applytagtoattrib tag1 attribs) bl1) ::
      embed_tags_end_attr attribs (Block(tag2, bl2) :: tl)
| Block (t, bl) :: tl ->
    let n_attribs = (applytagtoattrib t attribs) in
    Block (t, embed_tags_end_attr n_attribs bl) :: embed_tags_end_attr attribs tl
| hd :: tl -> hd :: embed_tags_end_attr attribs tl
let embed_tags = embed_tags_end_attr initial_attributes

(* Change <tag2><tag1>foo</tag1>bar</tag2>  to *)
(* <tag1>foo </tag1><tag2>bar</tag2>  *)
(* whenever tags don't commute and both are equivalent *)

let rec unembed_start_tags_attr attribs list = match list with
| [] -> []
| Block (t, bl) :: tl -> (match bl with
  | Block (tag2, bl2) :: tl2 ->
    (* can we unembed? *)
      let original = Block (t, [Block (tag2, bl2)])
      and trial = Block (tag2, bl2) in
      let ok =
	not (do_tags_commute t tag2) &&
	decorated_output_confirmed_match ([original], attribs)
	  ([trial], attribs) in
      if ok then
	unembed_start_tags_attr attribs (trial :: Block (t, tl2) :: tl)
      else
	let n_attribs = (applytagtoattrib t attribs) in 
	Block (t, (unembed_start_tags_attr n_attribs (Block (tag2, bl2) :: tl2)))
	:: unembed_start_tags_attr attribs tl
  | _ ->
      let n_attribs = (applytagtoattrib t attribs) in
      Block (t, unembed_start_tags_attr n_attribs bl) ::
      unembed_start_tags_attr attribs tl)
| hd :: tl -> hd :: unembed_start_tags_attr attribs tl
let unembed_start_tags = unembed_start_tags_attr initial_attributes

(* Change <tag1>foo <tag2>bar</tag2></tag1>  to *)
(* <tag1>foo </tag1><tag2>bar</tag2>  *)
(* whenever tags don't commute and both are equivalent *)

let rec unembed_end_tags_attr attribs list = match list with
| [] -> []
| Block (t, bl) :: tl -> (let rev_bl = List.rev bl in match rev_bl with
  | Block (tag2, bl2) :: tl2 ->
    (* can we unembed? *)
      let original = Block (t, [Block (tag2, bl2)])
      and trial = Block (tag2, bl2) in
      let ok =
	not (do_tags_commute t tag2) &&
	decorated_output_confirmed_match ([original], attribs)
	  ([trial], attribs) in
      if ok then
	unembed_end_tags_attr attribs (Block (t, List.rev tl2) :: trial :: tl)
      else
	let n_attribs = (applytagtoattrib t attribs) in
	Block (t, unembed_end_tags_attr n_attribs bl) ::
	unembed_end_tags_attr attribs tl
  | _ ->
      let n_attribs = (applytagtoattrib t attribs) in
      Block (t, unembed_end_tags_attr n_attribs bl) ::
      unembed_end_tags_attr attribs tl)
| hd :: tl -> hd :: unembed_end_tags_attr attribs tl
let unembed_end_tags = unembed_end_tags_attr initial_attributes

(* <tag1>foo</tag1><tag2>bar</tag2><tag1>baz</tag1> *)
(* -> <tag1>foo<tag2>bar</tag2>ba</tag1> *)
(* where tag2 destroys effect of tag1 *)

let rec includeable_blocks tag so_far attribs list = match list with
| Block (tag1, bl1) :: tl ->
    let original_block = Block (tag1, bl1) in
    let so_far = original_block :: so_far in
    if tag1 = tag then (* we've got it! *)
      (so_far, tl)
    else
      (* does applying the tag to the block wreck it? *)
      let n_attribs = (applytagtoattrib tag attribs) in 
      if decorated_output_confirmed_match ([original_block], attribs)
	  ([original_block], n_attribs) then
	includeable_blocks tag so_far attribs tl
      else
	([], [])
| _ -> ([], []) (* forget it -- none includeable *)


let rec widen_tag_scope_attr attribs list = match list with
| [] -> []
| Block (tag1, bl1) :: tl ->
    (* either find tag1 in tl or find non-block, or block that will change 
       when tag wraps it *)
    let (rev_includable, rest) = includeable_blocks tag1 [] attribs tl in
    (match rev_includable with
    | Block (tag2, bl2) :: tl2 -> (* do it! *)
	widen_tag_scope_attr attribs (Block (tag2, bl1 @ (List.rev_append tl2 bl2)) :: rest)
    | _ -> (* can't do it *) let n_attribs = (applytagtoattrib tag1 attribs) in 
      Block (tag1, widen_tag_scope_attr n_attribs bl1) ::
      widen_tag_scope_attr attribs tl
)
| hd :: tl -> hd :: widen_tag_scope_attr attribs tl

let widen_tag_scope = widen_tag_scope_attr initial_attributes;;

(* merge adjacent blocks if there are any tags in common *)
(* don't just look at the top tag; look at the "commuting" tag *)
(* that is, look at all nested tags that commute with the top *)

let rec commuting_tags_for_block so_far b = match b with
| [Block(tag, bl)] ->
    let rec all_commute l t =
      match l with
      | [] -> true
      | hd :: tl ->
	  if not (do_tags_commute t hd) then false else all_commute tl t in
    if all_commute so_far tag then commuting_tags_for_block (tag :: so_far) bl
    else so_far
| _ -> so_far

let rec contents_depth b n =
  if n == 0 then b else match b with
  | [Block(t1, bl)] -> contents_depth bl (n-1)
  | _ -> b
;;

let intersect_lists l1 l2 = List.filter (function x -> List.mem x l1) l2

let block_from_tag_list tags ignore_tag contents =
  let most_tags = List.filter (function x -> x != ignore_tag) tags in
  List.fold_left (function c -> function tag -> [Block(tag, c)])
    contents most_tags

let rec merge_adjacent_commutable attr list = match list with
| [] -> []
| Block (tag1, bl1) :: Block (tag2, bl2) :: tl ->
    let tags1 = commuting_tags_for_block [tag1] bl1
    and tags2 = commuting_tags_for_block [tag2] bl2 in
    (* if tags1 intersect tags2 is not empty, use that tag *)
    let common_tags = intersect_lists tags1 tags2 in
    if List.length common_tags == 0 then
      let n_attribs = (applytagtoattrib tag1 attr) in
      let rest = Block (tag2, bl2) :: tl in
      Block (tag1, merge_adjacent_commutable n_attribs bl1)
      :: (merge_adjacent_commutable attr rest)
    else
      let common_tag = List.hd common_tags in
      let sub1 = contents_depth bl1 ((List.length tags1)-1)
      and sub2 = contents_depth bl2 ((List.length tags2)-1) in
      let new_block_1 = block_from_tag_list tags1 common_tag sub1
      and new_block_2 = block_from_tag_list tags2 common_tag sub2 in
      let new_list = Block (common_tag, new_block_1 @ new_block_2) :: tl in
      merge_adjacent_commutable attr new_list
| Block (tag1, bl1) :: tl ->
      let n_attribs = (applytagtoattrib tag1 attr) in
      Block (tag1, merge_adjacent_commutable n_attribs bl1)
      :: (merge_adjacent_commutable attr tl)
| hd :: tl -> hd :: (merge_adjacent_commutable attr tl)
let merge_adjacent  = merge_adjacent_commutable initial_attributes
    


(* output file *)

let string_of_tag c = match c with
  'E' -> "EM" | 'P' -> "PL" | 'T' -> "TT" | _ -> String.make 1 c

let length_of_tag c = match c with
  'E' | 'P' | 'T' -> 2 | _ -> 1

let rec dump_block b = match b with
  Text t -> print_string t
|  Space s -> print_string s
|  Block (tag, blocklist) -> printf "<%s>" (string_of_tag tag);
    dump_block_list blocklist;
    printf "</%s>" (string_of_tag tag)
and dump_block_list b = List.iter dump_block b

(* return column *)
let pretty_out t space_pending column =
  if space_pending then
    let need_new_line = (String.length t) + column > 75 in 
    printf "%s%s" (if need_new_line then "\n" else " ") t;
    if need_new_line then String.length t else (String.length t) + column
  else let column = (String.length t) + column in
    print_string t; column

(* return (attribs, space_pending, column) *)

let rec dump_block_pretty l attribs space_pending column =
  match l with
  | [] -> (attribs, space_pending, column)
  | Text t :: rst ->
      let column = pretty_out t space_pending column in
      dump_block_pretty rst attribs false column
  | Space s :: rst ->
      if attribs.tt then
	let column = pretty_out s space_pending column in
	dump_block_pretty rst attribs false column
      else
	let column = pretty_out "" space_pending column in
	dump_block_pretty rst attribs true column
  | Block (tag, blocklist) :: rst ->
      let column = pretty_out (sprintf "<%s>" (string_of_tag tag)) space_pending column in
      let (attribs,space_pending,column) =
	dump_block_pretty blocklist (applytagtoattrib tag attribs) false column in
      let column = pretty_out (sprintf "</%s>" (string_of_tag tag)) space_pending column in
      dump_block_pretty rst attribs false column
let dump_pretty block =
  let (a,s,c) = dump_block_pretty block initial_attributes false 0 in
  if s then print_string "\n"

let rec cost_block subtotal b = match b with
  Text t -> subtotal + (String.length t)
|  Space s -> subtotal + (String.length s)
|  Block (tag, blocklist) ->
    2 * (length_of_tag tag) + 5 + (cost_blocklist subtotal blocklist);
and cost_blocklist subtotal b = List.fold_left cost_block subtotal b
let cost = cost_blocklist 0

let single_pass_optimizers x = 
  remove_useless (merge_whitespace (collapse_whitespace x))

let iterative_optimizers x =
  widen_tag_scope (
  remove_useless (
  embed_tags (
  unembed_start_tags ( 
  unembed_end_tags (
  remove_em_pairs (
  merge_adjacent (
  sort_tags (
  sort_spaces (
  x)))))))));;

let optimize r =
  let r = single_pass_optimizers r in
  let rec do_iterative_opt r =
    let startcost = cost r in
    let r = iterative_optimizers r in
    if startcost = (cost r) then r
    else do_iterative_opt r in
  do_iterative_opt r

let dumpboth r1 r2 =
  printf "*****\n";
  dump_block_list r1;
  printf "*****\n";
  dump_block_list r2;
  printf "*****\n"

(* show which tags commute *)
let alltags = [ '0'; '1'; 'r'; 'g'; 'b'; 'B'; 'I'; 'U'; 'T'; 'E'; 'S'; 'P' ];;
let show_comm ()=
  printf "    ";
  List.iter (function c -> printf "  %c" c) alltags;
  print_string "\n-----------------------------------------------------\n";
  List.iter (function t1 ->
    printf "%c  |" t1;
    List.iter (function x -> print_string (if do_tags_commute t1 x then "  o" else "  .")) alltags;
    print_newline ();) alltags
;;

let _ =
  let lexbuf = Lexing.from_channel stdin in
  let r = Parser.document Lexer.beginLex lexbuf in
  let s1 = merge_whitespace r in
  let s2 = optimize s1 in
  dump_block_list (if (decorated_output_confirmed_match (s1, initial_attributes) (s2, initial_attributes)) then s2 else s1)
(*    assert(decorated_output_confirmed_match (s1, initial_attributes) (s2, initial_attributes))*)
(*  let optimize = sort_spaces in*)
(*  let r = Parser.document Lexer.beginLex lexbuf in dump_block_list (optimize r);; *)
(*  let r = Parser.document Lexer.beginLex lexbuf in dump_block_list (remove_em_pairs (merge_whitespace (collapse_whitespace r)));; *)
